@@ -175,7 +175,7 @@
 #define TRACE_INIT_CREDS_SALT_PRINC(c, salt) \
     TRACE(c, (c, "Salt derived from principal: {data}", salt))
 #define TRACE_INIT_CREDS_SERVICE(c, service) \
-    TRACE(c, (c, "Setting initial creds service to {string}", service))
+    TRACE(c, (c, "Setting initial creds service to {str}", service))
 
 #define TRACE_KT_GET_ENTRY(c, keytab, princ, vno, enctype, err) \
     TRACE(c, (c, "Retrieving {princ} from {keytab} (vno {int}, " \
@@ -330,6 +330,9 @@
     TRACE(c, (c, "TGS request result: {kerr}", code))
 #define TRACE_TKT_CREDS_RETRY_TCP(c) \
     TRACE(c, (c, "Request or response is too big for UDP; retrying with TCP"))
+#define TRACE_TKT_CREDS_SAME_REALM_TGT(c, realm)                        \
+    TRACE(c, (c, "Received TGT referral back to same realm ({data}); trying " \
+              "again without referrals", realm))
 #define TRACE_TKT_CREDS_SERVICE_REQ(c, princ, referral) \
     TRACE(c, (c, "Requesting tickets for {princ}, referrals {str}", princ, \
               (referral) ? "on" : "off"))
