@@ -77,7 +77,6 @@ typedef struct gss_cred_id_struct {
 } gss_union_cred_desc, *gss_union_cred_t;
 
 typedef	OM_uint32	    (*gss_acquire_cred_with_password_sfct)(
-		    void *,		/* context */
 		    OM_uint32 *,	/* minor_status */
 		    const gss_name_t,	/* desired_name */
 		    const gss_buffer_t, /* password */
@@ -114,9 +113,9 @@ OM_uint32 gssint_get_mech_type_oid(gss_OID OID, gss_buffer_t token);
  *
  * This contants all of the functions defined in gssapi.h except for
  * gss_release_buffer() and gss_release_oid_set(), which I am
- * assuming, for now, to be equal across mechanisms.  
+ * assuming, for now, to be equal across mechanisms.
  */
- 
+
 typedef struct gss_config {
     gss_OID_desc    mech_type;
     void *	    context;
@@ -397,7 +396,7 @@ typedef struct gss_config {
 	OM_uint32       (*gssspi_set_cred_option)
     	(
     		    OM_uint32 *,	/* minor_status */
-    		    gss_cred_id_t,      /* cred_handle */
+    		    gss_cred_id_t *,    /* cred_handle */
     		    const gss_OID,      /* OID */
     		    const gss_buffer_t	/* value */
     		    );
@@ -473,11 +472,152 @@ typedef struct gss_config {
     		    gss_buffer_t	/* input_message_buffer */
     		    );
 
+	/* New for 1.8 */
+
+	OM_uint32	(*gss_acquire_cred_impersonate_name)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    const gss_cred_id_t,	/* impersonator_cred_handle */
+	    const gss_name_t,		/* desired_name */
+	    OM_uint32,			/* time_req */
+	    const gss_OID_set,		/* desired_mechs */
+	    gss_cred_usage_t,		/* cred_usage */
+	    gss_cred_id_t *,		/* output_cred_handle */
+	    gss_OID_set *,		/* actual_mechs */
+	    OM_uint32 *			/* time_rec */
+	/* */);
+
+	OM_uint32	(*gss_add_cred_impersonate_name)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_cred_id_t,		/* input_cred_handle */
+	    const gss_cred_id_t,	/* impersonator_cred_handle */
+	    const gss_name_t,		/* desired_name */
+	    const gss_OID,		/* desired_mech */
+	    gss_cred_usage_t,		/* cred_usage */
+	    OM_uint32,			/* initiator_time_req */
+	    OM_uint32,			/* acceptor_time_req */
+	    gss_cred_id_t *,		/* output_cred_handle */
+	    gss_OID_set *,		/* actual_mechs */
+	    OM_uint32 *,		/* initiator_time_rec */
+	    OM_uint32 *			/* acceptor_time_rec */
+	/* */);
+
+	OM_uint32	(*gss_display_name_ext)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    gss_OID,			/* display_as_name_type */
+	    gss_buffer_t		/* display_name */
+	/* */);
+
+	OM_uint32	(*gss_inquire_name)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    int *,			/* name_is_MN */
+	    gss_OID *,			/* MN_mech */
+	    gss_buffer_set_t *		/* attrs */
+	/* */);
+
+	OM_uint32	(*gss_get_name_attribute)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    gss_buffer_t,		/* attr */
+	    int *,			/* authenticated */
+	    int *,			/* complete */
+	    gss_buffer_t,		/* value */
+	    gss_buffer_t,		/* display_value */
+	    int *			/* more */
+	/* */);
+
+	OM_uint32	(*gss_set_name_attribute)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    int,			/* complete */
+	    gss_buffer_t,		/* attr */
+	    gss_buffer_t		/* value */
+	/* */);
+
+	OM_uint32	(*gss_delete_name_attribute)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    gss_buffer_t		/* attr */
+	/* */);
+
+	OM_uint32	(*gss_export_name_composite)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    gss_buffer_t		/* exp_composite_name */
+	/* */);
+
+	OM_uint32	(*gss_map_name_to_any)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    int,			/* authenticated */
+	    gss_buffer_t,		/* type_id */
+	    gss_any_t *			/* output */
+	/* */);
+
+	OM_uint32	(*gss_release_any_name_mapping)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_name_t,			/* name */
+	    gss_buffer_t,		/* type_id */
+	    gss_any_t *			/* input */
+	/* */);
+
+        OM_uint32       (*gss_pseudo_random)
+        (
+            OM_uint32 *,                /* minor_status */
+            gss_ctx_id_t,               /* context */
+            int,                        /* prf_key */
+            const gss_buffer_t,         /* prf_in */
+            ssize_t,                    /* desired_output_len */
+            gss_buffer_t                /* prf_out */
+        /* */);
+
+	OM_uint32	(*gss_set_neg_mechs)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_cred_id_t,		/* cred_handle */
+	    const gss_OID_set		/* mech_set */
+	/* */);
+
+	OM_uint32	(*gss_inquire_saslname_for_mech)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    const gss_OID,		/* desired_mech */
+	    gss_buffer_t,		/* sasl_mech_name */
+	    gss_buffer_t,		/* mech_name */
+	    gss_buffer_t		/* mech_description */
+	/* */);
+
+	OM_uint32	(*gss_inquire_mech_for_saslname)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    const gss_buffer_t,		/* sasl_mech_name */
+	    gss_OID *			/* mech_type */
+	/* */);
+
+	OM_uint32	(*gss_inquire_attrs_for_mech)
+	(
+	    OM_uint32 *,		/* minor_status */
+	    gss_const_OID,		/* mech */
+	    gss_OID_set *,		/* mech_attrs */
+	    gss_OID_set *		/* known_mech_attrs */
+	/* */);
+
 } *gss_mechanism;
 
 /* This structure MUST NOT be used by any code outside libgss */
 typedef struct gss_config_ext {
-    gss_acquire_cred_with_password_sfct	gss_acquire_cred_with_password;
+    gss_acquire_cred_with_password_sfct	gssspi_acquire_cred_with_password;
 } *gss_mechanism_ext;
 
 /*
@@ -587,18 +727,6 @@ gssint_userok(
 	int *			/* user_ok */
 );
 
-OM_uint32
-gss_store_cred(
-	OM_uint32 *,		/* minor_status */
-	const gss_cred_id_t,	/* input_cred_handle */
-	gss_cred_usage_t,	/* cred_usage */
-	const gss_OID,		/* desired_mech */
-	OM_uint32,		/* overwrite_cred */
-	OM_uint32,		/* default_cred */
-	gss_OID_set *,		/* elements_stored */
-	gss_cred_usage_t *	/* cred_usage_stored */
-);
-
 int
 gssint_get_der_length(
 	unsigned char **,	/* buf */
