@@ -25,7 +25,6 @@
  * or implied warranty.
  *
  *
- * krb5_get_in_tkt()
  */
 
 #include <string.h>
@@ -45,48 +44,8 @@
 #endif
 #endif /* APPLE_PKINIT */
 
-/*
-  All-purpose initial ticket routine, usually called via
-  krb5_get_in_tkt_with_password or krb5_get_in_tkt_with_skey.
-
-  Attempts to get an initial ticket for creds->client to use server
-  creds->server, (realm is taken from creds->client), with options
-  options, and using creds->times.starttime, creds->times.endtime,
-  creds->times.renew_till as from, till, and rtime.
-  creds->times.renew_till is ignored unless the RENEWABLE option is requested.
-
-  key_proc is called to fill in the key to be used for decryption.
-  keyseed is passed on to key_proc.
-
-  decrypt_proc is called to perform the decryption of the response (the
-  encrypted part is in dec_rep->enc_part; the decrypted part should be
-  allocated and filled into dec_rep->enc_part2
-  arg is passed on to decrypt_proc.
-
-  If addrs is non-NULL, it is used for the addresses requested.  If it is
-  null, the system standard addresses are used.
-
-  A succesful call will place the ticket in the credentials cache ccache
-  and fill in creds with the ticket information used/returned..
-
-  returns system errors, encryption errors
-
-*/
-
-
 /* some typedef's for the function args to make things look a bit cleaner */
 
-typedef krb5_error_code (*git_key_proc) (krb5_context,
-                                         krb5_enctype,
-                                         krb5_data *,
-                                         krb5_const_pointer,
-                                         krb5_keyblock **);
-
-typedef krb5_error_code (*git_decrypt_proc) (krb5_context,
-                                             const krb5_keyblock *,
-                                             krb5_const_pointer,
-                                             krb5_kdc_rep * );
-
 static krb5_error_code make_preauth_list (krb5_context,
                                           krb5_preauthtype *,
                                           int, krb5_pa_data ***);
@@ -112,175 +71,14 @@ krb5int_addint32 (krb5_int32 x, krb5_int32 y)
     return x + y;
 }
 
-#if APPLE_PKINIT
-/*
- * Common code to generate krb5_kdc_req.nonce. Like the original MIT code this
- * just uses krb5_timeofday(); it should use a PRNG. Even more unfortunately this
- * value is used interchangeably with an explicit now_time throughout this module...
- */
 static krb5_error_code
-gen_nonce(krb5_context  context,
-          krb5_int32    *nonce)
+decrypt_as_reply(krb5_context context, krb5_kdc_req *request,
+                 krb5_kdc_rep *as_reply, krb5_keyblock *key)
 {
-    krb5_int32 time_now;
-    krb5_error_code retval = krb5_timeofday(context, &time_now);
-    if(retval) {
-        return retval;
-    }
-    *nonce = time_now;
-    return 0;
-}
-#endif /* APPLE_PKINIT */
-
-/*
- * This function sends a request to the KDC, and gets back a response;
- * the response is parsed into ret_err_reply or ret_as_reply if the
- * reponse is a KRB_ERROR or a KRB_AS_REP packet.  If it is some other
- * unexpected response, an error is returned.
- */
-static krb5_error_code
-send_as_request(krb5_context            context,
-                krb5_data *packet, const krb5_data *realm,
-                krb5_error **           ret_err_reply,
-                krb5_kdc_rep **         ret_as_reply,
-                int                         *use_master)
-{
-    krb5_kdc_rep *as_reply = 0;
-    krb5_error_code retval;
-    krb5_data reply;
-    char k4_version;            /* same type as *(krb5_data::data) */
-    int tcp_only = 0;
-
-    reply.data = 0;
-
-    /* set the nonce if the caller expects us to do it */
-
-    k4_version = packet->data[0];
-send_again:
-    retval = krb5_sendto_kdc(context, packet,
-                             realm,
-                             &reply, use_master, tcp_only);
-#if APPLE_PKINIT
-    inTktDebug("krb5_sendto_kdc returned %d\n", (int)retval);
-#endif /* APPLE_PKINIT */
-
-    if (retval)
-        goto cleanup;
-
-    /* now decode the reply...could be error or as_rep */
-    if (krb5_is_krb_error(&reply)) {
-        krb5_error *err_reply;
-
-        if ((retval = decode_krb5_error(&reply, &err_reply)))
-            /* some other error code--??? */
-            goto cleanup;
-
-        if (ret_err_reply) {
-            if (err_reply->error == KRB_ERR_RESPONSE_TOO_BIG
-                && tcp_only == 0) {
-                tcp_only = 1;
-                krb5_free_error(context, err_reply);
-                free(reply.data);
-                reply.data = 0;
-                goto send_again;
-            }
-            *ret_err_reply = err_reply;
-        } else
-            krb5_free_error(context, err_reply);
-        goto cleanup;
-    }
-
-    /*
-     * Check to make sure it isn't a V4 reply.
-     */
-    if (!krb5_is_as_rep(&reply)) {
-/* these are in <kerberosIV/prot.h> as well but it isn't worth including. */
-#define V4_KRB_PROT_VERSION     4
-#define V4_AUTH_MSG_ERR_REPLY   (5<<1)
-        /* check here for V4 reply */
-        unsigned int t_switch;
-
-        /* From v4 g_in_tkt.c: This used to be
-           switch (pkt_msg_type(rpkt) & ~1) {
-           but SCO 3.2v4 cc compiled that incorrectly.  */
-        t_switch = reply.data[1];
-        t_switch &= ~1;
-
-        if (t_switch == V4_AUTH_MSG_ERR_REPLY
-            && (reply.data[0] == V4_KRB_PROT_VERSION
-                || reply.data[0] == k4_version)) {
-            retval = KRB5KRB_AP_ERR_V4_REPLY;
-        } else {
-            retval = KRB5KRB_AP_ERR_MSG_TYPE;
-        }
-        goto cleanup;
-    }
-
-    /* It must be a KRB_AS_REP message, or an bad returned packet */
-    if ((retval = decode_krb5_as_rep(&reply, &as_reply)))
-        /* some other error code ??? */
-        goto cleanup;
-
-    if (as_reply->msg_type != KRB5_AS_REP) {
-        retval = KRB5KRB_AP_ERR_MSG_TYPE;
-        krb5_free_kdc_rep(context, as_reply);
-        goto cleanup;
-    }
-
-    if (ret_as_reply)
-        *ret_as_reply = as_reply;
-    else
-        krb5_free_kdc_rep(context, as_reply);
-
-cleanup:
-    if (reply.data)
-        free(reply.data);
-    return retval;
-}
-
-static krb5_error_code
-decrypt_as_reply(krb5_context           context,
-                 krb5_kdc_req           *request,
-                 krb5_kdc_rep           *as_reply,
-                 git_key_proc           key_proc,
-                 krb5_const_pointer     keyseed,
-                 krb5_keyblock *        key,
-                 git_decrypt_proc       decrypt_proc,
-                 krb5_const_pointer     decryptarg)
-{
-    krb5_error_code             retval;
-    krb5_keyblock *             decrypt_key = 0;
-    krb5_data                   salt;
-
     if (as_reply->enc_part2)
         return 0;
 
-    if (key)
-        decrypt_key = key;
-    else {
-        /*
-         * Use salt corresponding to the client principal supplied by
-         * the KDC, which may differ from the requested principal if
-         * canonicalization is in effect.  We will check
-         * as_reply->client later in verify_as_reply.
-         */
-        if ((retval = krb5_principal2salt(context, as_reply->client, &salt)))
-            return(retval);
-
-        retval = (*key_proc)(context, as_reply->enc_part.enctype,
-                             &salt, keyseed, &decrypt_key);
-        free(salt.data);
-        if (retval)
-            goto cleanup;
-    }
-
-    if ((retval = (*decrypt_proc)(context, decrypt_key, decryptarg, as_reply)))
-        goto cleanup;
-
-cleanup:
-    if (!key && decrypt_key)
-        krb5_free_keyblock(context, decrypt_key);
-    return (retval);
+    return krb5_kdc_rep_decrypt_proc(context, key, NULL, as_reply);
 }
 
 /**
@@ -561,63 +359,6 @@ make_preauth_list(krb5_context  context,
 }
 
 #define MAX_IN_TKT_LOOPS 16
-static const krb5_enctype get_in_tkt_enctypes[] = {
-    ENCTYPE_DES3_CBC_SHA1,
-    ENCTYPE_ARCFOUR_HMAC,
-    ENCTYPE_DES_CBC_MD5,
-    ENCTYPE_DES_CBC_MD4,
-    ENCTYPE_DES_CBC_CRC,
-    0
-};
-
-static krb5_error_code
-rewrite_server_realm(krb5_context context,
-                     krb5_const_principal old_server,
-                     const krb5_data *realm,
-                     krb5_boolean tgs,
-                     krb5_principal *server)
-{
-    krb5_error_code retval;
-
-    assert(*server == NULL);
-
-    retval = krb5_copy_principal(context, old_server, server);
-    if (retval)
-        return retval;
-
-    krb5_free_data_contents(context, &(*server)->realm);
-    (*server)->realm.data = NULL;
-
-    retval = krb5int_copy_data_contents(context, realm, &(*server)->realm);
-    if (retval)
-        goto cleanup;
-
-    if (tgs) {
-        krb5_free_data_contents(context, &(*server)->data[1]);
-        (*server)->data[1].data = NULL;
-
-        retval = krb5int_copy_data_contents(context, realm, &(*server)->data[1]);
-        if (retval)
-            goto cleanup;
-    }
-
-cleanup:
-    if (retval) {
-        krb5_free_principal(context, *server);
-        *server = NULL;
-    }
-
-    return retval;
-}
-
-static inline int
-tgt_is_local_realm(krb5_creds *tgt)
-{
-    return (tgt->server->length == 2
-            && data_eq_string(tgt->server->data[0], KRB5_TGS_NAME)
-            && data_eq(tgt->server->data[1], tgt->client->realm)
-            && data_eq(tgt->server->realm, tgt->client->realm));
-}
 
 static krb5_error_code
 request_enc_pa_rep(krb5_pa_data ***padptr)
@@ -643,271 +384,6 @@ request_enc_pa_rep(krb5_pa_data ***padptr)
     return 0;
 }
 
-krb5_error_code KRB5_CALLCONV
-krb5_get_in_tkt(krb5_context context,
-                krb5_flags options,
-                krb5_address * const * addrs,
-                krb5_enctype * ktypes,
-                krb5_preauthtype * ptypes,
-                git_key_proc key_proc,
-                krb5_const_pointer keyseed,
-                git_decrypt_proc decrypt_proc,
-                krb5_const_pointer decryptarg,
-                krb5_creds * creds,
-                krb5_ccache ccache,
-                krb5_kdc_rep ** ret_as_reply)
-{
-    krb5_error_code     retval;
-    krb5_timestamp      time_now;
-    krb5_keyblock *     decrypt_key = 0;
-    krb5_kdc_req        request;
-    krb5_data *encoded_request;
-    krb5_error *        err_reply;
-    krb5_kdc_rep *      as_reply = 0;
-    krb5_pa_data  **    preauth_to_use = 0;
-    int                 loopcount = 0;
-    krb5_int32          do_more = 0;
-    int                 canon_flag;
-    int             use_master = 0;
-    int                 referral_count = 0;
-    krb5_principal_data referred_client;
-    krb5_principal      referred_server = NULL;
-    krb5_boolean        is_tgt_req;
-
-#if APPLE_PKINIT
-    inTktDebug("krb5_get_in_tkt top\n");
-#endif /* APPLE_PKINIT */
-
-    if (! krb5_realm_compare(context, creds->client, creds->server))
-        return KRB5_IN_TKT_REALM_MISMATCH;
-
-    if (ret_as_reply)
-        *ret_as_reply = 0;
-
-    referred_client = *(creds->client);
-    referred_client.realm.data = NULL;
-    referred_client.realm.length = 0;
-
-    /* per referrals draft, enterprise principals imply canonicalization */
-    canon_flag = ((options & KDC_OPT_CANONICALIZE) != 0) ||
-        creds->client->type == KRB5_NT_ENTERPRISE_PRINCIPAL;
-
-    /*
-     * Set up the basic request structure
-     */
-    request.magic = KV5M_KDC_REQ;
-    request.msg_type = KRB5_AS_REQ;
-    request.addresses = 0;
-    request.ktype = 0;
-    request.padata = 0;
-    if (addrs)
-        request.addresses = (krb5_address **) addrs;
-    else
-        if ((retval = krb5_os_localaddr(context, &request.addresses)))
-            goto cleanup;
-    request.kdc_options = options;
-    request.client = creds->client;
-    request.server = creds->server;
-    request.nonce = 0;
-    request.from = creds->times.starttime;
-    request.till = creds->times.endtime;
-    request.rtime = creds->times.renew_till;
-#if APPLE_PKINIT
-    retval = gen_nonce(context, (krb5_int32 *)&time_now);
-    if(retval) {
-        goto cleanup;
-    }
-    request.nonce = time_now;
-#endif /* APPLE_PKINIT */
-
-    retval = krb5int_copy_etypes(get_in_tkt_enctypes, &request.ktype);
-    request.nktypes = krb5int_count_etypes(request.ktype);
-    if (ktypes) {
-        int i, req, next = 0;
-        for (req = 0; ktypes[req]; req++) {
-            if (ktypes[req] == request.ktype[next]) {
-                next++;
-                continue;
-            }
-            for (i = next + 1; i < request.nktypes; i++)
-                if (ktypes[req] == request.ktype[i]) {
-                    /* Found the enctype we want, but not in the
-                       position we want.  Move it, but keep the old
-                       one from the desired slot around in case it's
-                       later in our requested-ktypes list.  */
-                    krb5_enctype t;
-                    t = request.ktype[next];
-                    request.ktype[next] = request.ktype[i];
-                    request.ktype[i] = t;
-                    next++;
-                    break;
-                }
-            /* If we didn't find it, don't do anything special, just
-               drop it.  */
-        }
-        request.ktype[next] = 0;
-        request.nktypes = next;
-    }
-    request.authorization_data.ciphertext.length = 0;
-    request.authorization_data.ciphertext.data = 0;
-    request.unenc_authdata = 0;
-    request.second_ticket = 0;
-
-    /*
-     * If a list of preauth types are passed in, convert it to a
-     * preauth_to_use list.
-     */
-    if (ptypes) {
-        retval = make_preauth_list(context, ptypes, -1, &preauth_to_use);
-        if (retval)
-            goto cleanup;
-    }
-
-    is_tgt_req = tgt_is_local_realm(creds);
-
-    while (1) {
-        if (loopcount++ > MAX_IN_TKT_LOOPS) {
-            retval = KRB5_GET_IN_TKT_LOOP;
-            goto cleanup;
-        }
-
-#if APPLE_PKINIT
-        inTktDebug("krb5_get_in_tkt calling krb5_obtain_padata\n");
-#endif /* APPLE_PKINIT */
-        if ((retval = krb5_obtain_padata(context, preauth_to_use, key_proc,
-                                         keyseed, creds, &request)) != 0)
-            goto cleanup;
-        if (preauth_to_use)
-            krb5_free_pa_data(context, preauth_to_use);
-        preauth_to_use = 0;
-
-        err_reply = 0;
-        as_reply = 0;
-
-        if ((retval = krb5_timeofday(context, &time_now)))
-            goto cleanup;
-
-        /*
-         * XXX we know they are the same size... and we should do
-         * something better than just the current time
-         */
-        request.nonce = (krb5_int32) time_now;
-
-        if ((retval = encode_krb5_as_req(&request, &encoded_request)) != 0)
-            goto cleanup;
-        retval = send_as_request(context, encoded_request,
-                                 krb5_princ_realm(context, request.client), &err_reply,
-                                 &as_reply, &use_master);
-        krb5_free_data(context, encoded_request);
-        if (retval != 0)
-            goto cleanup;
-
-        if (err_reply) {
-            if (err_reply->error == KDC_ERR_PREAUTH_REQUIRED &&
-                err_reply->e_data.length > 0) {
-                retval = decode_krb5_padata_sequence(&err_reply->e_data,
-                                                     &preauth_to_use);
-                krb5_free_error(context, err_reply);
-                if (retval)
-                    goto cleanup;
-                retval = sort_krb5_padata_sequence(context,
-                                                   &request.server->realm,
-                                                   preauth_to_use);
-                if (retval)
-                    goto cleanup;
-                continue;
-            } else if (canon_flag && err_reply->error == KDC_ERR_WRONG_REALM) {
-                if (++referral_count > KRB5_REFERRAL_MAXHOPS ||
-                    err_reply->client == NULL ||
-                    err_reply->client->realm.length == 0) {
-                    retval = KRB5KDC_ERR_WRONG_REALM;
-                    krb5_free_error(context, err_reply);
-                    goto cleanup;
-                }
-                /* Rewrite request.client with realm from error reply */
-                if (referred_client.realm.data) {
-                    krb5_free_data_contents(context, &referred_client.realm);
-                    referred_client.realm.data = NULL;
-                }
-                retval = krb5int_copy_data_contents(context,
-                                                    &err_reply->client->realm,
-                                                    &referred_client.realm);
-                krb5_free_error(context, err_reply);
-                if (retval)
-                    goto cleanup;
-                request.client = &referred_client;
-
-                if (referred_server != NULL) {
-                    krb5_free_principal(context, referred_server);
-                    referred_server = NULL;
-                }
-
-                retval = rewrite_server_realm(context,
-                                              creds->server,
-                                              &referred_client.realm,
-                                              is_tgt_req,
-                                              &referred_server);
-                if (retval)
-                    goto cleanup;
-                request.server = referred_server;
-
-                continue;
-            } else {
-                retval = (krb5_error_code) err_reply->error
-                    + ERROR_TABLE_BASE_krb5;
-                krb5_free_error(context, err_reply);
-                goto cleanup;
-            }
-        } else if (!as_reply) {
-            retval = KRB5KRB_AP_ERR_MSG_TYPE;
-            goto cleanup;
-        }
-        if ((retval = krb5_process_padata(context, &request, as_reply,
-                                          key_proc, keyseed, decrypt_proc,
-                                          &decrypt_key, creds,
-                                          &do_more)) != 0)
-            goto cleanup;
-
-        if (!do_more)
-            break;
-    }
-
-    if ((retval = decrypt_as_reply(context, &request, as_reply, key_proc,
-                                   keyseed, decrypt_key, decrypt_proc,
-                                   decryptarg)))
-        goto cleanup;
-
-    if ((retval = verify_as_reply(context, time_now, &request, as_reply)))
-        goto cleanup;
-
-    if ((retval = stash_as_reply(context, time_now, &request, as_reply,
-                                 creds, ccache)))
-        goto cleanup;
-
-cleanup:
-    if (request.ktype)
-        free(request.ktype);
-    if (!addrs && request.addresses)
-        krb5_free_addresses(context, request.addresses);
-    if (request.padata)
-        krb5_free_pa_data(context, request.padata);
-    if (preauth_to_use)
-        krb5_free_pa_data(context, preauth_to_use);
-    if (decrypt_key)
-        krb5_free_keyblock(context, decrypt_key);
-    if (as_reply) {
-        if (ret_as_reply)
-            *ret_as_reply = as_reply;
-        else
-            krb5_free_kdc_rep(context, as_reply);
-    }
-    if (referred_client.realm.data)
-        krb5_free_data_contents(context, &referred_client.realm);
-    if (referred_server)
-        krb5_free_principal(context, referred_server);
-    return (retval);
-}
-
 /* Sort a pa_data sequence so that types named in the "preferred_preauth_types"
  * libdefaults entry are listed before any others. */
 static krb5_error_code
@@ -992,6 +468,11 @@ build_in_tkt_name(krb5_context context,
     *server = NULL;
 
     if (in_tkt_service) {
+        /* Minimally invasive fix for inability to change password with no
+         * default realm, for backporting. */
+        if (strcmp(in_tkt_service, "kadmin/changepw") == 0)
+            in_tkt_service = "kadmin/changepw@";
+
         /* this is ugly, because so are the data structures involved.  I'm
            in the library, so I'm going to manipulate the data structures
            directly, otherwise, it will be worse. */
@@ -1092,9 +573,10 @@ init_creds_get(krb5_context context,
                                     &request,
                                     &realm,
                                     &flags);
-        if (code == KRB5KRB_ERR_RESPONSE_TOO_BIG && !tcp_only)
+        if (code == KRB5KRB_ERR_RESPONSE_TOO_BIG && !tcp_only) {
+            TRACE_INIT_CREDS_RETRY_TCP(context);
             tcp_only = 1;
-        else if (code != 0 || (flags & 1) == 0)
+        } else if (code != 0 || !(flags & KRB5_INIT_CREDS_STEP_FLAG_CONTINUE))
             break;
 
         krb5_free_data_contents(context, &reply);
@@ -1130,7 +612,7 @@ krb5_init_creds_get_creds(krb5_context context,
                           krb5_init_creds_context ctx,
                           krb5_creds *creds)
 {
-    if ((ctx->flags & KRB5_INIT_CREDS_STEP_FLAG_COMPLETE) == 0)
+    if (!ctx->complete)
         return KRB5_NO_TKT_SUPPLIED;
 
     return krb5int_copy_creds_contents(context, &ctx->cred, creds);
@@ -1141,7 +623,7 @@ krb5_init_creds_get_times(krb5_context context,
                           krb5_init_creds_context ctx,
                           krb5_ticket_times *times)
 {
-    if ((ctx->flags & KRB5_INIT_CREDS_STEP_FLAG_COMPLETE) == 0)
+    if (!ctx->complete)
         return KRB5_NO_TKT_SUPPLIED;
 
     *times = ctx->cred.times;
@@ -1321,6 +803,9 @@ krb5_init_creds_init(krb5_context context,
     int tmp;
     char *str = NULL;
     krb5_gic_opt_ext *opte;
+    krb5_get_init_creds_opt local_opts;
+
+    TRACE_INIT_CREDS(context, client);
 
     ctx = k5alloc(sizeof(*ctx), &code);
     if (code != 0)
@@ -1343,9 +828,14 @@ krb5_init_creds_init(krb5_context context,
     ctx->start_time = start_time;
 
     if (options == NULL) {
-        code = krb5_get_init_creds_opt_alloc(context, &options);
-        if (code != 0)
-            goto cleanup;
+        /*
+         * We initialize a non-extended options because that way the shadowed
+         * flag will be sent and they will be freed when the init_creds context
+         * is freed. The options will be extended and copied off the stack into
+         * storage by opt_to_opte.
+         */
+        krb5_get_init_creds_opt_init(&local_opts);
+        options = &local_opts;
     }
 
     code = krb5int_gic_opt_to_opte(context, options,
@@ -1520,6 +1010,8 @@ krb5_init_creds_set_service(krb5_context context,
 {
     char *s;
 
+    TRACE_INIT_CREDS_SERVICE(context, service);
+
     s = strdup(service);
     if (s == NULL)
         return ENOMEM;
@@ -1553,6 +1045,8 @@ init_creds_validate_reply(krb5_context context,
 
         assert(error != NULL);
 
+        TRACE_INIT_CREDS_ERROR_REPLY(context,
+                                     error->error + ERROR_TABLE_BASE_krb5);
         if (error->error == KRB_ERR_RESPONSE_TOO_BIG) {
             krb5_free_error(context, error);
             return KRB5KRB_ERR_RESPONSE_TOO_BIG;
@@ -1692,6 +1186,8 @@ init_creds_step_request(krb5_context context,
         goto cleanup;
 
 cleanup:
+    krb5_free_pa_data(context, ctx->request->padata);
+    ctx->request->padata = NULL;
     return code;
 }
 
@@ -1734,12 +1230,18 @@ static krb5_boolean
 negotiation_requests_restart(krb5_context context, krb5_init_creds_context ctx,
                              krb5_pa_data **padata)
 {
-    if (!ctx->have_restarted &&
-        (krb5int_upgrade_to_fast_p(context, ctx->fast_state, padata) ||
-         (ctx->err_reply->error == KDC_ERR_PREAUTH_FAILED &&
-          !ctx->sent_nontrivial_preauth)))
-        return 1;
-    return 0;
+    if (ctx->have_restarted)
+        return FALSE;
+    if (krb5int_upgrade_to_fast_p(context, ctx->fast_state, padata)) {
+        TRACE_INIT_CREDS_RESTART_FAST(context);
+        return TRUE;
+    }
+    if (ctx->err_reply->error == KDC_ERR_PREAUTH_FAILED &&
+        !ctx->sent_nontrivial_preauth) {
+        TRACE_INIT_CREDS_RESTART_PREAUTH_FAILED(context);
+        return TRUE;
+    }
+    return FALSE;
 }
 
 /* Ensure that the reply enctype was among the requested enctypes. */
@@ -1813,6 +1315,7 @@ init_creds_step_reply(krb5_context context,
                 code = KRB5KDC_ERR_WRONG_REALM;
                 goto cleanup;
             }
+            TRACE_INIT_CREDS_REFERRAL(context, &ctx->err_reply->client->realm);
             /* Rewrite request.client with realm from error reply */
             krb5_free_data_contents(context, &ctx->request->client->realm);
             code = krb5int_copy_data_contents(context,
@@ -1895,6 +1398,7 @@ init_creds_step_reply(krb5_context context,
      */
     if (ctx->salt.length == SALT_TYPE_AFS_LENGTH && ctx->salt.data == NULL) {
         code = krb5_principal2salt(context, ctx->reply->client, &ctx->salt);
+        TRACE_INIT_CREDS_SALT_PRINC(context, &ctx->salt);
         if (code != 0)
             goto cleanup;
     }
@@ -1912,18 +1416,20 @@ init_creds_step_reply(krb5_context context,
        as_key at all yet, then use the gak_fct to get one, and try
        again.  */
     if (ctx->as_key.length) {
+        TRACE_INIT_CREDS_AS_KEY_PREAUTH(context, &ctx->as_key);
         code = krb5int_fast_reply_key(context, strengthen_key, &ctx->as_key,
                                       &encrypting_key);
         if (code != 0)
             goto cleanup;
-        code = decrypt_as_reply(context, NULL, ctx->reply, NULL, NULL,
-                                &encrypting_key, krb5_kdc_rep_decrypt_proc,
-                                NULL);
+        code = decrypt_as_reply(context, NULL, ctx->reply, &encrypting_key);
+        if (code != 0)
+            TRACE_INIT_CREDS_PREAUTH_DECRYPT_FAIL(context, code);
     } else
         code = -1;
 
     if (code != 0) {
         /* if we haven't get gotten a key, get it now */
+        TRACE_INIT_CREDS_GAK(context, &ctx->salt, &ctx->s2kparams);
         code = (*ctx->gak_fct)(context, ctx->request->client,
                                ctx->reply->enc_part.enctype,
                                ctx->prompter, ctx->prompter_data,
@@ -1931,19 +1437,20 @@ init_creds_step_reply(krb5_context context,
                                &ctx->as_key, ctx->gak_data);
         if (code != 0)
             goto cleanup;
+        TRACE_INIT_CREDS_AS_KEY_GAK(context, &ctx->as_key);
 
         code = krb5int_fast_reply_key(context, strengthen_key, &ctx->as_key,
                                       &encrypting_key);
         if (code != 0)
             goto cleanup;
 
-        code = decrypt_as_reply(context, NULL, ctx->reply, NULL, NULL,
-                                &encrypting_key, krb5_kdc_rep_decrypt_proc,
-                                NULL);
+        code = decrypt_as_reply(context, NULL, ctx->reply, &encrypting_key);
         if (code != 0)
             goto cleanup;
     }
 
+    TRACE_INIT_CREDS_DECRYPTED_REPLY(context, ctx->reply->enc_part2->session);
+
     code = krb5int_fast_verify_nego(context, ctx->fast_state,
                                     ctx->reply, ctx->encoded_previous_request,
                                     &encrypting_key, &fast_avail);
@@ -1991,7 +1498,7 @@ init_creds_step_reply(krb5_context context,
 
     /* success */
     code = 0;
-    ctx->flags |= KRB5_INIT_CREDS_STEP_FLAG_COMPLETE;
+    ctx->complete = TRUE;
 
 cleanup:
     krb5_free_pa_data(context, padata);
@@ -2026,8 +1533,8 @@ krb5_init_creds_step(krb5_context context,
     realm->data = NULL;
     realm->length = 0;
 
-    if (ctx->flags & KRB5_INIT_CREDS_STEP_FLAG_COMPLETE)
-        goto cleanup;
+    if (ctx->complete)
+        return EINVAL;
 
     if (in->length != 0) {
         code = init_creds_step_reply(context, ctx, in);
@@ -2041,7 +1548,7 @@ krb5_init_creds_step(krb5_context context,
             }
             goto copy_realm;
         }
-        if (code != 0 || (ctx->flags & KRB5_INIT_CREDS_STEP_FLAG_COMPLETE))
+        if (code != 0 || ctx->complete)
             goto cleanup;
     }
 
@@ -2077,8 +1584,7 @@ cleanup:
         }
     }
 
-    *flags = (ctx->flags & KRB5_INIT_CREDS_STEP_FLAG_COMPLETE) ? 0 : 1;
-
+    *flags = ctx->complete ? 0 : KRB5_INIT_CREDS_STEP_FLAG_CONTINUE;
     return code;
 }
 
