@@ -52,7 +52,7 @@
  */
 #define KRB5_MAJOR_RELEASE 1
 #define KRB5_MINOR_RELEASE 9
-#define KRB5_PATCHLEVEL 3
-/* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20120206"
-#define KRB5_RELTAG "tags/krb5-1-9-3-final"
+#define KRB5_PATCHLEVEL 4
+/* #undef  KRB5_RELTAIL */
+#define KRB5_RELDATE "20120620"
+#define KRB5_RELTAG "krb5-1.9.4-final"
