@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2005 Massachusetts Institute of Technology
+ * Copyright (c) 2007 Secure Endpoints Inc.
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -91,8 +92,8 @@ kmm_load_pending(void) {
 
 /*! \internal
   \brief Message handler for the registrar thread. */
-khm_boolean KHMAPI kmmint_reg_cb(khm_int32 msg_type, 
-                                 khm_int32 msg_sub_type, 
+khm_boolean KHMAPI kmmint_reg_cb(khm_int32 msg_type,
+                                 khm_int32 msg_sub_type,
                                  khm_ui_4 uparam,
                                  void *vparam)
 {
@@ -131,6 +132,9 @@ khm_boolean KHMAPI kmmint_reg_cb(khm_int32 msg_type,
   callback routine ( kmmint_reg_cb() ) */
 DWORD WINAPI kmmint_registrar(LPVOID lpParameter)
 {
+
+    PDESCTHREAD(L"KMM Registrar", L"KMM");
+
     tid_registrar = GetCurrentThreadId();
 
     kmq_subscribe(KMSG_KMM, kmmint_reg_cb);
@@ -140,9 +144,10 @@ DWORD WINAPI kmmint_registrar(LPVOID lpParameter)
 
     while(KHM_SUCCEEDED(kmq_dispatch(INFINITE)));
 
+    kmq_unsubscribe(KMSG_KMM, kmmint_reg_cb);
+    kmq_unsubscribe(KMSG_SYSTEM, kmmint_reg_cb);
+
     ExitThread(0);
-    /* not reached */
-    return 0;
 }
 
 /*! \internal
@@ -156,6 +161,8 @@ DWORD WINAPI kmmint_plugin_broker(LPVOID lpParameter)
     DWORD rv = 0;
     kmm_plugin_i * p = (kmm_plugin_i *) lpParameter;
 
+    PDESCTHREAD(p->p.name, L"KMM");
+
     _begin_task(0);
     _report_mr1(KHERR_NONE, MSG_PB_START, _cstr(p->p.name));
     _describe();
@@ -166,14 +173,9 @@ DWORD WINAPI kmmint_plugin_broker(LPVOID lpParameter)
 
     p->tid_thread = GetCurrentThreadId();
 
-    if (IsBadCodePtr(p->p.msg_proc)) {
-        _report_mr0(KHERR_WARNING, MSG_PB_INVALID_CODE_PTR);
-        rv = KHM_ERROR_INVALID_PARAM;
-    } else {
-        rv = (*p->p.msg_proc)(KMSG_SYSTEM, KMSG_SYSTEM_INIT, 
-                              0, (void *) &(p->p));
-        _report_mr1(KHERR_INFO, MSG_PB_INIT_RV, _int32(rv));
-    }
+    rv = (*p->p.msg_proc)(KMSG_SYSTEM, KMSG_SYSTEM_INIT,
+                          0, (void *) &(p->p));
+    _report_mr1(KHERR_INFO, MSG_PB_INIT_RV, _int32(rv));
 
     /* if it fails to initialize, we exit the plugin */
     if(KHM_FAILED(rv)) {
@@ -280,15 +282,12 @@ DWORD WINAPI kmmint_plugin_broker(LPVOID lpParameter)
         p->state = KMM_PLUGIN_STATE_EXITED;
 
     /* the following call will automatically release the plugin */
-    kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, 
+    kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG,
                      KMM_REG_EXIT_PLUGIN, (void *) p);
 
     TlsSetValue(tls_kmm, (LPVOID) 0);
 
     ExitThread(rv);
-
-    /* not reached */
-    return rv;
 }
 
 /*! \internal
@@ -324,8 +323,13 @@ void kmmint_init_plugin(kmm_plugin_i * p) {
            from the initial attempt to start the plugin.  Undo
            the hold we did a few lines earlier. */
         kmm_release_plugin(kmm_handle_from_plugin(p));
+
         /* same for the plugin count for the module. */
+#ifdef DEBUG
+        assert(p->flags & KMM_PLUGIN_FLAG_IN_MODCOUNT);
+#endif
         p->module->plugin_count--;
+        p->flags &= ~KMM_PLUGIN_FLAG_IN_MODCOUNT;
     }
 
     p->state = KMM_PLUGIN_STATE_PREINIT;
@@ -342,32 +346,23 @@ void kmmint_init_plugin(kmm_plugin_i * p) {
         goto _exit;
     }
 
-    if(KHM_FAILED(kmm_get_plugin_config(p->p.name, 0, &csp_plugin)) ||
-       KHM_FAILED(khc_read_int32(csp_plugin, L"Flags", &t))) {
+    if(KHM_FAILED(kmm_get_plugin_config(p->p.name, 0, &csp_plugin))) {
         if(KHM_FAILED(kmm_register_plugin(&(p->p), 0))) {
             _report_mr0(KHERR_ERROR, MSG_IP_NOT_REGISTERED);
 
             p->state = KMM_PLUGIN_STATE_FAIL_NOT_REGISTERED;
             goto _exit;
         }
-        
+
         if(KHM_FAILED(kmm_get_plugin_config(p->p.name, 0, &csp_plugin))) {
             _report_mr0(KHERR_ERROR, MSG_IP_NOT_REGISTERED);
 
             p->state = KMM_PLUGIN_STATE_FAIL_NOT_REGISTERED;
             goto _exit;
         }
-
-        if(KHM_FAILED(khc_read_int32(csp_plugin, L"Flags", &t))) {
-            _report_mr0(KHERR_ERROR, MSG_IP_NOT_REGISTERED);
-
-            p->state = KMM_PLUGIN_STATE_FAIL_NOT_REGISTERED;
-            goto _exit;
-        }
-
     }
 
-    if(t & KMM_PLUGIN_FLAG_DISABLED) {
+    if (KHM_SUCCEEDED(khc_read_int32(csp_plugin, L"Disabled", &t)) && t) {
         p->flags |= KMM_PLUGIN_FLAG_DISABLED;
         p->state = KMM_PLUGIN_STATE_FAIL_DISABLED;
         goto _exit;
@@ -383,18 +378,18 @@ void kmmint_init_plugin(kmm_plugin_i * p) {
 
     p->n_depends = 0;
     p->n_unresolved = 0;
-    
+
     do {
         wchar_t * deps = NULL;
         wchar_t * d;
         khm_size sz = 0;
 
-        if(khc_read_multi_string(csp_plugin, L"Dependencies", 
+        if(khc_read_multi_string(csp_plugin, L"Dependencies",
                                  NULL, &sz) != KHM_ERROR_TOO_LONG)
             break;
 
         deps = PMALLOC(sz);
-        if(KHM_FAILED(khc_read_multi_string(csp_plugin, L"Dependencies", 
+        if(KHM_FAILED(khc_read_multi_string(csp_plugin, L"Dependencies",
                                             deps, &sz))) {
             if(deps)
                 PFREE(deps);
@@ -444,6 +439,10 @@ void kmmint_init_plugin(kmm_plugin_i * p) {
 
     } while(FALSE);
 
+#ifdef DEBUG
+    assert(!(p->flags & KMM_PLUGIN_FLAG_IN_MODCOUNT));
+#endif
+    p->flags |= KMM_PLUGIN_FLAG_IN_MODCOUNT;
     p->module->plugin_count++;
 
     LeaveCriticalSection(&cs_kmm);
@@ -474,11 +473,11 @@ _exit_post:
     if(csp_plugins != NULL)
         khc_close_space(csp_plugins);
 
-    _report_mr2(KHERR_INFO, MSG_IP_STATE, 
+    _report_mr2(KHERR_INFO, MSG_IP_STATE,
                 _dupstr(p->p.name), _int32(p->state));
 
     _end_task();
-    
+
     return;
 
     /* jump here if an error condition happens before the plugin
@@ -490,13 +489,22 @@ _exit:
     if(csp_plugins != NULL)
         khc_close_space(csp_plugins);
 
-    _report_mr2(KHERR_WARNING, MSG_IP_EXITING, 
+    _report_mr2(KHERR_WARNING, MSG_IP_EXITING,
                 _dupstr(p->p.name), _int32(p->state));
     _end_task();
 
+
+#ifdef ASYNC_PLUGIN_UNLOAD_ON_FAILURE
+
     kmm_hold_plugin(kmm_handle_from_plugin(p));
 
     kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, KMM_REG_EXIT_PLUGIN, (void *) p);
+
+#else
+
+    kmmint_exit_plugin(p);
+
+#endif
 }
 
 /*! \internal
@@ -504,7 +512,7 @@ _exit:
 
   In addition to terminating the thread, and removing p from the
   linked list and hashtable, it also frees up p.
-   
+
   \note Should only be called from the context of the registrar thread. */
 void kmmint_exit_plugin(kmm_plugin_i * p) {
     int np;
@@ -531,21 +539,23 @@ void kmmint_exit_plugin(kmm_plugin_i * p) {
     EnterCriticalSection(&cs_kmm);
 
     /* undo reference count done in kmmint_init_plugin() */
-    if(p->state == KMM_PLUGIN_STATE_EXITED ||
-       p->state == KMM_PLUGIN_STATE_HOLD) {
+    if(p->flags & KMM_PLUGIN_FLAG_IN_MODCOUNT) {
 
         np = --(p->module->plugin_count);
+        p->flags &= ~KMM_PLUGIN_FLAG_IN_MODCOUNT;
 
     } else {
-        /* the plugin was never active.  We can't base a module unload
-           decision on np */
+        /* the plugin was not included in the module's plugin_count.
+           We can't base a decision to unload the module based on this
+           plugin exiting. */
         np = TRUE;
     }
 
     /* The plugin is in an error state.  We need to keep the plugin
-       record in tact so that the failure information is kept
-       around. */
-    if (p->state < KMM_PLUGIN_STATE_NONE) {
+       record intact so that the failure information is kept around.
+       Also, we shouldn't release the plugin if it appears that
+       kmmint_init_plugin() was never called for it. */
+    if (p->state < KMM_PLUGIN_STATE_HOLD) {
         release_plugin = FALSE;
     }
 
@@ -594,6 +604,13 @@ void kmmint_init_module(kmm_module_i * m) {
 
     kmm_hold_module(kmm_handle_from_module(m));
 
+    /* If the module is not in the pre-init state, we can't
+       initialize it. */
+    if(m->state != KMM_MODULE_STATE_PREINIT) {
+        _report_mr1(KHERR_INFO, MSG_IM_NOT_PREINIT, _int32(m->state));
+        goto _exit;
+    }
+
     if(KHM_FAILED(kmm_get_modules_config(0, &csp_mods))) {
         _report_mr0(KHERR_ERROR, MSG_IM_GET_CONFIG);
         _location(L"kmm_get_modules_config()");
@@ -605,13 +622,6 @@ void kmmint_init_module(kmm_module_i * m) {
     khc_read_int32(csp_mods, L"ModuleMaxFailureCount", &max_fail_count);
     khc_read_int64(csp_mods, L"ModuleFailureCountResetTime", &fail_reset_time);
 
-    /* If the module is not in the pre-init state, we can't
-       initialize it. */
-    if(m->state != KMM_MODULE_STATE_PREINIT) {
-        _report_mr1(KHERR_INFO, MSG_IM_NOT_PREINIT, _int32(m->state));
-        goto _exit;
-    }
-
     if(KHM_FAILED(kmm_get_module_config(m->name, 0, &csp_mod))) {
         _report_mr0(KHERR_ERROR, MSG_IM_NOT_REGISTERED);
 
@@ -619,15 +629,17 @@ void kmmint_init_module(kmm_module_i * m) {
         goto _exit;
     }
 
-    if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"Flags", &i)) &&
-       (i & KMM_MODULE_FLAG_DISABLED)) {
-
+    if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"Disabled", &i)) && i) {
         _report_mr0(KHERR_INFO, MSG_IM_DISABLED);
 
         m->state = KMM_MODULE_STATE_FAIL_DISABLED;
         goto _exit;
     }
 
+    if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"NoUnload", &i)) && i) {
+        m->flags |= KMM_MODULE_FLAG_NOUNLOAD;
+    }
+
     if(KHM_SUCCEEDED(khc_read_int32(csp_mod, L"FailureCount", &i))) {
         khm_int64 tm;
         khm_int64 ct;
@@ -642,7 +654,7 @@ void kmmint_init_module(kmm_module_i * m) {
 
         ct = (FtToInt(&fct) - tm) / 10000000i64;
 
-        if(tm > 0 && 
+        if(tm > 0 &&
            ct > fail_reset_time) {
             i = 0;
             khc_write_int32(csp_mod, L"FailureCount", 0);
@@ -654,7 +666,7 @@ void kmmint_init_module(kmm_module_i * m) {
         /* did we exceed the max failure count?  However, we ignore
            the max failure count if the reason why it didn't load the
            last time was because the module wasn't found. */
-        if(i > max_fail_count && 
+        if(i > max_fail_count &&
            last_reason != KMM_MODULE_STATE_FAIL_NOT_FOUND) {
             /* failed too many times */
             _report_mr0(KHERR_INFO, MSG_IM_MAX_FAIL);
@@ -664,14 +676,18 @@ void kmmint_init_module(kmm_module_i * m) {
         }
     }
 
-    if(khc_read_string(csp_mod, L"ImagePath", NULL, &sz) == 
+    if(khc_read_string(csp_mod, KMM_VALNAME_IMAGEPATH, NULL, &sz) ==
        KHM_ERROR_TOO_LONG) {
         if(m->path)
             PFREE(m->path);
         m->path = PMALLOC(sz);
-        khc_read_string(csp_mod, L"ImagePath", m->path, &sz);
+        khc_read_string(csp_mod, KMM_VALNAME_IMAGEPATH, m->path, &sz);
     } else {
-        _report_mr0(KHERR_ERROR, MSG_IM_NOT_REGISTERED);
+	/*
+	 * If there is no image path, then the module has not been
+	 * installed.  Do not report an error and bother the user.
+	 *   _report_mr0(KHERR_ERROR, MSG_IM_NOT_REGISTERED);
+	 */
 
         m->state = KMM_MODULE_STATE_FAIL_NOT_REGISTERED;
         goto _exit;
@@ -815,7 +831,7 @@ void kmmint_init_module(kmm_module_i * m) {
     if(csp_mods)
         khc_close_space(csp_mods);
 
-    _report_mr2(KHERR_INFO, MSG_IM_MOD_STATE, 
+    _report_mr2(KHERR_INFO, MSG_IM_MOD_STATE,
                 _dupstr(m->name), _int32(m->state));
 
     kmmint_remove_from_module_queue();
@@ -884,7 +900,7 @@ void kmmint_init_module(kmm_module_i * m) {
 void kmmint_exit_module(kmm_module_i * m) {
     kmm_plugin_i * p;
 
-    /*  Exiting a module happens in two stages.  
+    /*  Exiting a module happens in two stages.
 
         If the module state is running (there are active plugins) then
         those plugins must be exited.  This has to be done from the
@@ -922,16 +938,23 @@ void kmmint_exit_module(kmm_module_i * m) {
         p = kmm_listed_plugins;
 
         while(p) {
-            if(p->module == m) {
+            if(p->module == m &&
+               (p->flags & KMM_PLUGIN_FLAG_IN_MODCOUNT)) {
+
                 kmm_hold_plugin(kmm_handle_from_plugin(p));
-                kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG, 
+                kmq_post_message(KMSG_KMM, KMSG_KMM_I_REG,
                                  KMM_REG_EXIT_PLUGIN, (void *) p);
                 np++;
+
             }
 
             p = LNEXT(p);
         }
 
+#ifdef DEBUG
+        assert(np == m->plugin_count);
+#endif
+
         if(np > 0) {
             /*  we have to go back and wait for the plugins to exit.
                 when the last plugin exits, it automatically posts
@@ -940,6 +963,21 @@ void kmmint_exit_module(kmm_module_i * m) {
             LeaveCriticalSection(&cs_kmm);
             return;
         }
+
+    } else {
+
+#ifdef DEBUG
+        assert(m->plugin_count == 0 ||
+               m->state == KMM_MODULE_STATE_EXITPLUG);
+#endif
+
+        /* if there are still plug-ins waiting to be unloaded, then we
+           have to go back and wait for them to finish.  Once they are
+           done, kmmint_exit_module() will get called again. */
+        if (m->plugin_count > 0) {
+            LeaveCriticalSection(&cs_kmm);
+            return;
+        }
     }
 
     if(m->flags & KMM_MODULE_FLAG_INITP) {
@@ -948,8 +986,8 @@ void kmmint_exit_module(kmm_module_i * m) {
         if(m->state > 0)
             m->state = KMM_MODULE_STATE_EXIT;
 
-        p_exit_module = 
-            (exit_module_t) GetProcAddress(m->h_module, 
+        p_exit_module =
+            (exit_module_t) GetProcAddress(m->h_module,
                                            EXP_EXIT_MODULE);
         if(p_exit_module) {
             LeaveCriticalSection(&cs_kmm);
@@ -963,16 +1001,26 @@ void kmmint_exit_module(kmm_module_i * m) {
 
     LeaveCriticalSection(&cs_kmm);
 
-    if(m->h_module) {
+    if(!(m->flags & KMM_MODULE_FLAG_NOUNLOAD) &&
+       m->h_module) {
         FreeLibrary(m->h_module);
     }
 
-    if(m->h_resource && (m->h_resource != m->h_module)) {
+    if(!(m->flags & KMM_MODULE_FLAG_NOUNLOAD) &&
+       m->h_resource && (m->h_resource != m->h_module)) {
         FreeLibrary(m->h_resource);
     }
 
     m->h_module = NULL;
     m->h_resource = NULL;
+
+    if (m->flags & KMM_MODULE_FLAG_LOADED) {
+#ifdef DEBUG
+        assert(kmm_active_modules > 0);
+#endif
+        kmm_active_modules--;
+    }
+
     m->flags = 0;
 
     /* release the hold obtained in kmmint_init_module() */
@@ -980,13 +1028,6 @@ void kmmint_exit_module(kmm_module_i * m) {
 
     /* Last but not least, now see if there are any modules left that
        are running. If not, we can safely signal an exit. */
-
-#ifdef DEBUG
-    assert(kmm_active_modules > 0);
-#endif
-
-    kmm_active_modules--;
-
     if (kmm_active_modules == 0) {
         SetEvent(evt_exit);
     }
